% Name server node
%
% state = {nodes, map, requests, mode}
% nodes = {NODES, TUNNELS, CHANGED}
% map = [{NAME, ID, ADDR}, ...]
% requests =  REQUESTS
% mode = 'master' | 'remote' | 'noidle'
%
% NODES is just a list of addresses.
%
% REQUESTS is a list of {PEER, SRC, RNODE} tuples holding
% requests for peers that could not be resolved yet.
%
% CHANGED is 'yes' if nodes where added or removed from NODES since
% the last idle-token loop was started.
%
% usage: strandvm [OPTION ...] dns -- [-remote|-noidle] [COUNT]

-exports(['$start'/0]).

'$start' :- 
    '$events'(ES), 
    command_line(ARGS),
    parse_args(ARGS, MODE),
    initial_state(MODE, ST),
    init_idle_token(MODE, [], []),
    loop(ES, ST).

parse_args([], MODE) :- MODE := master.
parse_args(['-remote'|_], MODE) :- MODE := remote.
parse_args(['-noidle'|_], MODE) :- MODE := noidle.
parse_args([_], MODE) :- MODE := master.
parse_args(_, _, _) :-
    otherwise | 
    error(usage('strandvm OPTIONS ... ns -- [-remote|-noidle] [MINCOUNT]')).

initial_state(MODE, ST) :- 
    ST := {{[], [], no}, [], [], MODE}.

% handle ns-specific events
loop([], _) :- '$log'('event stream closed').
loop([register(NODE)|ES], {{ALL, TUNS, _}, MAP, RS, MODE}) :-
    register_node(ALL, NODE, ALL2),
    propagate_node(TUNS, register_remote(NODE)),
    do_pending_requests(RS, {{ALL2, TUNS, yes}, MAP, RS, MODE}, ES).
loop([register_remote(NODE)|ES], {{ALL, TUNS, _}, MAP, RS, MODE}) :-
    % not propagated
    register_node(ALL, NODE, ALL2),
    do_pending_requests(RS, {{ALL2, TUNS, yes}, MAP, RS, MODE}, ES).
loop([register(NODE, NAME)|ES], {{ALL, TUNS, _}, MAP, RS, MODE}) :-
    register_node(ALL, NODE, ALL2),
    register_name(MAP, NAME, NODE, MAP2),
    propagate_node(TUNS, register_remote(NODE, NAME)),
    do_pending_requests(RS, {{ALL2, TUNS, yes}, MAP2, RS, MODE}, ES).
loop([register_remote(NODE, NAME)|ES], {{ALL, TUNS, _}, MAP, RS, MODE}) :-
    % not propagated
    register_node(ALL, NODE, ALL2),
    register_name(MAP, NAME, NODE, MAP2),
    do_pending_requests(RS, {{ALL2, TUNS, yes}, MAP2, RS, MODE}, ES).
loop([register_tunnel(NODE)|ES], {{ALL, TUNS, _}, MAP, RS, MODE}) :-
    register_node(TUNS, NODE, TUNS2),
    RNS is NODE << 8,
    propagate_all_nodes(ALL, RNS),
    loop(ES, {{ALL, TUNS2, yes}, MAP, RS, MODE}).
loop([lookup(PEER, RNODE, SRC)|ES], ST) :-
    '$log'(lookup(PEER, SRC)),
    lookup_peer(PEER, SRC, RNODE, ST, ST2),
    loop(ES, ST2).
loop([idle_token(TOK)|ES], ST) :-
    handle_idle_token(TOK, ST, ST2),
    loop(ES, ST2).
loop([idle_broadcast|ES], ST) :-
    broadcast_idle(ST),
    loop(ES, ST).
loop([E|ES], ST) :- handle(E, ST), loop(ES, ST).

handle_idle_token(_, {{ALL, TUNS, _}, MAP, RS, noidle}, ST) :-
    ST := {{ALL, TUNS, no}, MAP, RS, noidle}.
handle_idle_token(_, {{ALL, TUNS, yes}, MAP, RS, master}, ST) :-
    init_idle_token(master, ALL, TUNS),
    ST := {{ALL, TUNS, no}, MAP, RS, master}.
handle_idle_token(TOK, ST1, ST) :-
    '$set_idle_token'(TOK),
    ST := ST1.

init_idle_token(noidle, _, _).
init_idle_token(remote, _, TUNS) :-
    length(TUNS, TN),    % ignore received register_tunnel/1 messages
    '$init_idle_token'([], 0, TN).
init_idle_token(master, NODES, TUNS) :-
    length(TUNS, TN),    % ignore received register_tunnel/1 messages
    ns_list(TUNS, L, NODES),
    '$node'(SELF),
    list_to_tuple([SELF|L], NT),
    length(NT, PN),
    RA is PN + TN - 1,  % ignore sent idle_token/1 messages between nodes
    '$init_idle_token'({NT, 1, 0, 0}, 0, RA).  % {NODES, INDEX, SUM, CYCLES}

ns_list([], L, TL) :- L := TL.
ns_list([T|TUNS], L, TL) :-
    NS is T << 8,
    L := [NS|L2],
    ns_list(TUNS, L2, TL).

broadcast_idle({{ALL, _, _}, _, _, _}) :-
    broadcast(ALL, idle, _).

% internal events
handle(send_read(ID, PID), _) :-
    '$node'(SELF), 
    send_with(SELF, PID, read(ID, PID, SELF)).
handle(send_value(ID, PID, X), _) :-
    '$send'(PID, value(ID, PID, X)).
handle(send_addref(PID, ID), _) :-
    '$send'(PID, addref(ID)).
% external events
handle(read(ID, SELF, PID), _) :-
    '$get_exposed'(ID, X), 
    send_value(X, PID, ID, SELF).  % suspends on X
handle(value(ID, OWNER, X), _) :- 
    '$assign_remote'(ID, OWNER, X).
handle(addref(ID), _) :- '$add_reference'(ID).
handle(nodes(NODES), {_, _, {ALL, _}, _}) :- 
    NODES := ALL.
handle(shutdown, {{ALL, TUNS, _}, _, _, _}) :-
    shutdown(ALL, TUNS).
handle(send_idle_token(TOK), _) :-
    get_arg(1, TOK, NODES),
    get_arg(2, TOK, INDEX),
    get_arg(INDEX, NODES, PID),
    '$send'(PID, idle_token(TOK)).
handle(terminate, _) :- halt(0).
handle(E, _) :- 
    otherwise | '$log'(unkown_event(E)).

send_with(X, DEST, MSG) :- 
    data(X) | '$send'(DEST, MSG).

send_value(X, PID, ID, SELF) :-
    data(X) |
    '$drop_reference'(ID),
    '$send'(PID, value(ID, SELF, X)).

do_pending_requests([], ST, ES) :- loop(ES, ST).
do_pending_requests([{_, _, R}|RS], ST, ES) :-
    known(R) |
    do_pending_requests(RS, ST, ES).
do_pending_requests([{PEER, NODE, R}|RS], ST, ES) :-
    unknown(R) |
    lookup_peer(PEER, NODE, R, ST, ST2),
    do_pending_requests(RS, ST2, ES).

lookup_peer(PEER, NODE, R, {NS, MAP, RS, MODE}, ST) :-
    lookup_peer2(MAP, PEER, NODE, R, {NS, MAP, RS, MODE}, 
        NMAP, NMAP, ST).

lookup_peer2([], PEER, NODE, R, {NS, _, RS, MODE}, NMAP, NMAPTL, 
    ST) :-
    NMAPTL := [],
    add_peer_request(RS, PEER, NODE, R, RS2),
    ST := {NS, NMAP, RS2, MODE},
    '$log'(peer_request_pending(PEER, NODE)).    
lookup_peer2([{PEER, ID, ADDR}|MAP], PEER, NODE, R, ST, NMAP, NMAPTL, 
    ST2) :-
    '$message_port_owner'(ADDR, ID2),
    lookup_peer3(ID, ID2, PEER, MAP, ADDR, NODE, R, ST, NMAP, NMAPTL, 
        ST2).
lookup_peer2([{PEER1, ID, ADDR}|MAP], PEER, NODE, R, ST, NMAP, NMAPTL, 
    ST2) :-
    PEER1 =\= PEER |
    NMAPTL := [{PEER, ID, ADDR}|NMAPTL2],
    lookup_peer2(MAP, PEER, NODE, R, ST, NMAP, NMAPTL2, ST2).

lookup_peer3(ID, ID, PEER, MAP, ADDR, NODE, R, {NS, _, RS, MODE}, 
    NMAP, NMAPTL, ST2) :-
    R := ADDR,
    NMAPTL := [{PEER, ID, ADDR}|MAP],
    ST2 := {NS, NMAP, RS, MODE},
    '$log'(peer_found(PEER, ADDR, NODE)),
    '$send'(NODE, peer(PEER, ADDR)).
lookup_peer3(ID1, ID, PEER, MAP, ADDR, NODE, R, {NS, _, RS, MODE}, 
    NMAP, NMAPTL, ST2) :-
    ID1 =\= ID |
    NMAPTL := MAP,
    '$log'(unowned_port(PEER, ID, ADDR)),
    lookup_peer2([], PEER, NODE, R, {NS, NMAP, RS, MODE}, NMAP, _, ST2).

register_node([], NODE, ALL) :-
    '$log'(register(NODE)),
    ALL := [NODE].
register_node([NODE|ALL], NODE, ALL2) :-
    '$log'(re_register(NODE)),
    ALL2 := [NODE|ALL].
register_node([NODE1|ALL], NODE, ALL2) :-
    otherwise |
    ALL2 := [NODE1|ALL3],
    register_node(ALL, NODE, ALL3).

register_name([], NAME, NODE, MAP) :-
    '$log'(register_name(NAME, NODE)),
    '$message_port_owner'(NODE, ID),
    MAP := [{NAME, ID, NODE}].
register_name([{NAME, _, ADDR}|MAP], NAME, NODE, MAP2) :-
    '$log'(re_register_name(NODE)),
    '$message_port_owner'(ADDR, ID),
    MAP2 := [{NAME, ID, ADDR}|MAP].
register_name([NODE1|MAP], NAME, NODE, MAP2) :-
    otherwise |
    MAP2 := [NODE1|MAP3],
    register_name(MAP, NAME, NODE, MAP3).

propagate_node([], _).
propagate_node([T|TUNS], MSG) :-
    RNS is T << 8,
    '$send'(RNS, MSG),
    propagate_node(TUNS, MSG).

propagate_all_nodes([], _).
propagate_all_nodes([N|ALL], RNS) :-
    '$send'(RNS, register_remote(N)),
    propagate_all_nodes(ALL, RNS).

broadcast(NS, MSG, DONE) :-
    broadcast(NS, [], MSG, DONE).

broadcast([], [], _, DONE) :- DONE := [].
broadcast([N|ALL], [], MSG, DONE) :-
    '$message_port_owner'(N, ID),
    broadcast2(ID, N, ALL, MSG, DONE).

broadcast2(0, _, ALL, MSG, DONE) :-
    broadcast(ALL, [], MSG, DONE).
broadcast2(ID, N, ALL, MSG, DONE) :-
    ID =\= 0 |
    '$send'(N, MSG, DONE1),
    broadcast(ALL, DONE1, MSG, DONE).

add_peer_request([], PEER, NODE, R, RS) :-
    RS := [{PEER, NODE, R}].
add_peer_request([{P, N, R}|RS], P, N, R2, RS2) :-
    RS2 := [{P, N, R}|RS],
    R2 := R.
add_peer_request([{P1, N1, R1}|RS], P, N, R, RS2) :-
    P =\= P1 |
    RS2 := [{P1, N1, R1}|RS3],
    add_peer_request(RS, P, N, R, RS3).

shutdown(ALL, TUNS) :- 
    remote_ns(TUNS, NS, ALL),
    broadcast(NS, silence, DONE1),
    broadcast(NS, DONE1, terminate, DONE2),
    '$node'(SELF),
    LMAC is SELF >> 8,
    terminate_tunnels(TUNS, DONE2, LMAC).

terminate_tunnels([], [], _) :- halt(0).
terminate_tunnels([T|TUNS], [], LMAC) :-
    TA is (LMAC << 8) + T,
    '$send'(TA, terminate, DONE),
    terminate_tunnels(TUNS, DONE, LMAC).

remote_ns([], R, TL) :- R := TL.
remote_ns([T|TUNS], R, TL) :-
    RNS is T << 8,
    R := [RNS|R2],
    remote_ns(TUNS, R2, TL).
