% Name/topology server node
%
% state = {nodes, topologies, requests, mode}
% nodes = {NODES, TUNNELS, COUNT}
% topologies = {MINCOUNT, RING, TORUS}
% requests =  {ALL, REQUESTS}
% mode = 'master' | 'remote' | 'noidle'
%
% NODES is just a list of addresses.
% RING/TORUS look like this:
%
%   [[NODE1, [PEER1|NODE2], ...], ...]
%
% REQUESTS is a list of {PEER, SRC, RNODE} tuples holding
% requests for peers that could not be resolved yet.
%
% COUNT gives the number of nodes registered, MINCOUNT is
% the number of nodes which must be registered before
% peer requests are processed.
%
% ALL is instantiated to NODES, once C >= MC.
%
% usage: strandvm [OPTION ...] ns -- [-remote|-noidle] COUNT

-exports(['$start'/0]).

'$start' :- 
    '$events'(ES), 
    command_line(ARGS),
    parse_args(ARGS, MINCOUNT, MODE),
    initial_state(MINCOUNT, MODE, ST),
    loop(ES, ST).

parse_args([], MC, MODE) :- 
    MC := 0, MODE := master.
parse_args(['-remote', MCOUNT], MC, MODE) :-
    string_to_integer(MCOUNT, MC),
    MODE := remote.
parse_args(['-noidle', MCOUNT], MC, MODE) :-
    string_to_integer(MCOUNT, MC),
    MODE := noidle.
parse_args([MCOUNT], MC, MODE) :-
    string_to_integer(MCOUNT, MC),
    MODE := master.
parse_args(_, _, _) :-
    otherwise | 
    error(usage('strandvm OPTIONS ... ns -- [-remote|-noidle] MINCOUNT')).    

initial_state(MC, MODE, ST) :- 
    ST := {{[], [], 0}, {MC, [], []}, {_, []}, MODE}.

% handle ns-specific events
loop([], _) :- '$log'('event stream closed').
loop([register(NODE)|ES], {{ALL, TUNS, C}, TOPS, RS, MODE}) :-
    register_node(ALL, NODE, local, ALL2),
    C1 is C + 1,
    propagate_node(TUNS, register_remote(NODE)),
    do_pending_requests({{ALL2, TUNS, C1}, TOPS, RS, MODE}, ES).
loop([register_remote(NODE)|ES], {{ALL, TUNS, C}, TOPS, RS, MODE}) :-
    % not propagated
    register_node(ALL, NODE, remote, ALL2),
    C1 is C + 1,
    do_pending_requests({{ALL2, TUNS, C1}, TOPS, RS, MODE}, ES).
loop([register_tunnel(NODE)|ES], {{ALL, TUNS, C}, TOPS, RS, MODE}) :-
    register_node(TUNS, NODE, tunnel, TUNS2),
    RNS is NODE << 8,
    C1 is C + 1,
    propagate_all_nodes(ALL, RNS),
    loop(ES, {{ALL, TUNS2, C1}, TOPS, RS, MODE}).
loop([topology(TYPE, TOP)|ES], ST) :-
    update_topology(TYPE, TOP, ST, ST2),
    loop(ES, ST2).
loop([lookup(PEER, RNODE, SRC)|ES], ST) :-
    counts(ST, C, MC),
    perform_lookup(PEER, RNODE, SRC, ES, C, MC, ST).
loop([idle_token(TOK)|ES], ST) :-
    '$set_idle_token'(TOK),
    loop(ES, ST).
loop([idle_broadcast|ES], ST) :-
    broadcast_idle(ST),
    loop(ES, ST).
loop([E|ES], ST) :- handle(E, ST), loop(ES, ST).

perform_lookup(PEER, RNODE, SRC, ES, C, MC, {NS, TOPS, {ALL,
    RS}, MODE}) :-
    C < MC |
    '$log'(lookup_postponed(PEER, SRC, C, MC)),
    add_peer_request(RS, PEER, SRC, RNODE, RS2),
    loop(ES, {NS, TOPS, {ALL, RS2}, MODE}).
perform_lookup(PEER, RNODE, SRC, ES, C, MC, ST) :-
    C >= MC |
    '$log'(lookup(PEER, SRC)),
    lookup_peer(PEER, SRC, RNODE, ST, ST2),
    loop(ES, ST2).

% process pending events for resolvable peer
do_pending_requests(ST, ES) :-
    counts(ST, C, MC),
    do_pending_requests1(C, MC, ST, ES).

do_pending_requests1(C, MC, ST, ES) :-
    C < MC |
    loop(ES, ST).
do_pending_requests1(C, MC, {NS, TOPS, {ALL, RS}, MODE}, ES) :-
    C >= MC |
    set_all(NS, ALL, MODE),
    do_pending_requests2(RS, {NS, TOPS, {ALL, []}, MODE}, ES).

do_pending_requests2([], ST, ES) :- loop(ES, ST).
do_pending_requests2([{PEER, NODE, R}|RS], ST, ES) :-
    lookup_peer(PEER, NODE, R, ST, ST2),
    do_pending_requests2(RS, ST2, ES).

set_all({NODES, TUNS, _}, ALL, MODE) :- 
    unknown(ALL) | 
    ALL := NODES, 
    init_idle_token(MODE, NODES, TUNS).
set_all(_, _, _) :- otherwise | true.

init_idle_token(noidle, _, _).
init_idle_token(remote, _, TUNS) :-
    length(TUNS, TN),    % ignore received register_tunnel/1 messages
    '$init_idle_token'([], 0, TN).
init_idle_token(master, NODES, TUNS) :-
    length(TUNS, TN),    % ignore received register_tunnel/1 messages
    ns_list(TUNS, L, NODES),
    '$node'(SELF),
    list_to_tuple([SELF|L], NT),
    length(NT, PN),
    RA is PN + TN - 1,  % ignore sent idle_token/1 messages between nodes
    '$init_idle_token'({NT, 1, 0, 0}, 0, RA).  % {NODES, INDEX, SUM, CYCLES}

ns_list([], L, TL) :- L := TL.
ns_list([T|TUNS], L, TL) :-
    NS is T << 8,
    L := [NS|L2],
    ns_list(TUNS, L2, TL).

broadcast_idle({{ALL, _, _}, _, _, _}) :-
    broadcast(ALL, idle, _).

% internal events
handle(send_read(ID, PID), _) :-
    '$node'(SELF), 
    send_with(SELF, PID, read(ID, PID, SELF)).
handle(send_value(ID, PID, X), _) :-
    '$send'(PID, value(ID, PID, X)).
handle(send_addref(PID, ID), _) :-
    '$send'(PID, addref(ID)).
% external events
handle(read(ID, SELF, PID), _) :-
    '$get_exposed'(ID, X), 
    send_value(X, PID, ID, SELF).  % suspends on X
handle(value(ID, OWNER, X), _) :- 
    '$assign_remote'(ID, OWNER, X).
handle(addref(ID), _) :- '$add_reference'(ID).
handle(nodes(NODES), {_, _, {ALL, _}, _}) :- 
    NODES := ALL.
handle(shutdown, {{_, TUNS, _}, _, {ALL, _}, _}) :-
    shutdown(ALL, TUNS).
handle(send_idle_token(TOK), _) :-
    get_arg(1, TOK, NODES),
    get_arg(2, TOK, INDEX),
    get_arg(INDEX, NODES, PID),
    '$send'(PID, idle_token(TOK)).
handle(terminate, _) :- halt(0).
handle(E, _) :- 
    otherwise | '$log'(unkown_event(E)).

send_with(X, DEST, MSG) :- 
    data(X) | '$send'(DEST, MSG).

send_value(X, PID, ID, SELF) :-
    data(X) |
    '$drop_reference'(ID),
    '$send'(PID, value(ID, SELF, X)).

lookup_peer(fwd, NODE, R, ST, ST2) :-
    lookup_ring(fwd, NODE, R, ST, ST2).
lookup_peer(bwd, NODE, R, ST, ST2) :-
    lookup_ring(bwd, NODE, R, ST, ST2).
lookup_peer(north, NODE, R, ST, ST2) :-
    lookup_torus(north, NODE, R, ST, ST2).
lookup_peer(east, NODE, R, ST, ST2) :-
    lookup_torus(east, NODE, R, ST, ST2).
lookup_peer(south, NODE, R, ST, ST2) :-
    lookup_torus(south, NODE, R, ST, ST2).
lookup_peer(west, NODE, R, ST, ST2) :-
    lookup_torus(west, NODE, R, ST, ST2).
lookup_peer(PEER, _, _, ST, ST2) :-
    otherwise |
    '$log'(unknown_peer(PEER)),
    ST2 := ST.

lookup_ring(PEER, NODE, R, {{ALL, TUNS, C}, {MC, [], TT},
    RS, MODE}, ST) :-
    topology:ring(ALL, RT),
    propagate_topology(TUNS, ring, RT),
    lookup_peer2(RT, PEER, NODE, R, {{ALL, TUNS, C},
        {MC, RT, TT}, RS, MODE}, ST).
lookup_ring(PEER, NODE, R, ST, ST2) :-
    otherwise |
    topology(ring, ST, RING),
    lookup_peer2(RING, PEER, NODE, R, ST, ST2).

lookup_torus(PEER, NODE, R, {{ALL, TUNS, C},
    {MC, RT, []}, RS, MODE}, ST) :-
    topology:torus(ALL, TT),
    propagate_topology(TUNS, torus, TT),
    lookup_peer2(TT, PEER, NODE, R, {{ALL, TUNS, C},
        {MC, RT, TT}, RS, MODE}, ST).
lookup_torus(PEER, NODE, R, ST, ST2) :-
    otherwise |
    topology(torus, ST, TORUS),
    lookup_peer2(TORUS, PEER, NODE, R, ST, ST2).

lookup_peer2([], PEER, NODE, R, {NS, TOPS, {ALL, RS}, MODE}, ST) :-
    add_peer_request(RS, PEER, NODE, R, RS2),
    ST := {NS, TOPS, {ALL, RS2}, MODE},
    '$log'(peer_request_pending(PEER, NODE)).    
lookup_peer2([[NODE|PEERS]|_], PEER, NODE, R, ST, ST2) :-
    lookup_peer_by_name(PEERS, PEER, NODE, R, no, ST, ST2).
lookup_peer2([[NODE1|_]|ALL], PEER, NODE, R, ST, ST2) :-
    NODE1 =\= NODE |
    lookup_peer2(ALL, PEER, NODE, R, ST, ST2).

lookup_peer_by_name([], PEER, NODE, R, yes, {NS, TOPS, {ALL, 
    RS}, MODE}, ST) :-
    add_peer_request(RS, PEER, NODE, R, RS2),
    ST := {NS, TOPS, {ALL, RS2}, MODE},
    '$log'(peer_request_pending(PEER, NODE)).    
lookup_peer_by_name([], PEER, NODE, _, no, ST, ST2) :-
    ST2 := ST,
    '$log'(no_peer_for_node(PEER, NODE)).    
lookup_peer_by_name([[PEER|NODE2]|_], PEER, NODE, R, _, ST, ST2) :-
    R := NODE2,
    ST2 := ST,
    '$log'(peer_found(PEER, NODE2, NODE)),
    '$send'(NODE, peer(PEER, NODE2)).
lookup_peer_by_name([[PEER1|_]|PEERS], PEER, NODE, R, M, 
    ST, ST2) :-
    PEER1 =\= PEER |
    lookup_peer_by_name(PEERS, PEER, NODE, R, M, ST, ST2).

% accessors
topology(ring, {_, {_, RING, _}, _, _}, T) :- T := RING.

topology(torus, {_, {_, _, TORUS}, _, _}, T) :- T := TORUS.

all_nodes({{ALL, _, _}, _, _, _}, R) :- R := ALL.

tunnels({{_, TUNS, _}, _, _, _}, R) :- R := TUNS.

counts({{_, _, C}, {MC, _, _}, _, _}, CR, MCR) :-
    CR := C, MCR := MC.

update_topology(ring, RT, {NS, {MC, _, TT}, RS, MODE}, ST2) :-
    ST2 := {NS, {MC, RT, TT}, RS, MODE}.
update_topology(torus, TT, {NS, {MC, RT, _}, RS, MODE}, ST2) :-
    ST2 := {NS, {MC, RT, TT}, RS, MODE}.

register_node([], NODE, TYPE, ALL) :-
    '$log'(register(TYPE, NODE)),
    ALL := [NODE].
register_node([NODE|ALL], NODE, TYPE, ALL2) :-  % needed at all?
    '$log'(re_register(TYPE, NODE)),
    ALL2 := [NODE|ALL].
register_node([NODE1|ALL], NODE, TYPE, ALL2) :-
    otherwise |
    ALL2 := [NODE1|ALL3],
    register_node(ALL, NODE, TYPE, ALL3).

propagate_node([], _).
propagate_node([T|TUNS], MSG) :-
    RNS is T << 8,
    '$send'(RNS, MSG),
    propagate_node(TUNS, MSG).

propagate_all_nodes([], _).
propagate_all_nodes([N|ALL], RNS) :-
    '$send'(RNS, register_remote(N)),
    propagate_all_nodes(ALL, RNS).

propagate_topology(TUNS, TYPE, TOP) :-
    deref(TOP, OK),
    propagate_topology2(TUNS, OK, topology(TYPE, TOP)).

propagate_topology2([], _, _).
propagate_topology2([T|TUNS], [], TN) :-
    RNS is T * 256,
    '$send'(RNS, TN),
    propagate_topology2(TUNS, [], TN).

broadcast(NS, MSG, DONE) :-
    broadcast(NS, [], MSG, DONE).

broadcast([], [], _, DONE) :- DONE := [].
broadcast([N|ALL], [], MSG, DONE) :-
    '$send'(N, MSG, DONE1),
    broadcast(ALL, DONE1, MSG, DONE).

add_peer_request([], PEER, NODE, R, RS) :-
    RS := [{PEER, NODE, R}].
add_peer_request([{P, N, R}|RS], P, N, R2, RS2) :-
    RS2 := [{P, N, R}|RS],
    R2 := R.
add_peer_request([{P1, N1, R1}|RS], P, N, R, RS2) :-
    otherwise |
    RS2 := [{P1, N1, R1}|RS3],
    add_peer_request(RS, P, N, R, RS3).

shutdown(ALL, TUNS) :- 
    remote_ns(TUNS, NS, ALL),
    broadcast(NS, silence, DONE1),
    broadcast(NS, DONE1, terminate, DONE2),
    '$node'(SELF),
    LMAC is SELF >> 8,
    terminate_tunnels(TUNS, DONE2, LMAC).

terminate_tunnels([], [], _) :- halt(0).
terminate_tunnels([T|TUNS], [], LMAC) :-
    TA is (LMAC << 8) + T,
    '$send'(TA, terminate, DONE),
    terminate_tunnels(TUNS, DONE, LMAC).

remote_ns([], R, TL) :- R := TL.
remote_ns([T|TUNS], R, TL) :-
    RNS is T << 8,
    R := [RNS|R2],
    remote_ns(TUNS, R2, TL).
