% system glue library for inter-node communication
%
% init
%   Starts event handling (done automatically when loaded from 
%   command line).
%
% handle(EVENT)
%   Handle single standard-protocol event.
%
% self(ADDR^)
%   Assign own address (machine- and port-identification) to ADDR.
%
% nodes(ALL^)
%   returns a list of all node addresses in the namespace.
%
% terminate
%   terminate all registered nodes, and stop name server.
%
% register(NAME?)
%   register at dynamic name-server under NAME (requires using a
%   "dns" name server node).

-exports(['$start'/0, init/0, handle/1, nodes/1, terminate/0, self/1, register/1]).

'$start' :- init.

init :-
    '$events'(ES), 
    '$node'(SELF),
    send_with(SELF, 0, register(SELF)),
    loop(ES).

loop([]) :- '$log'('event stream closed').
loop([E|ES]) :- handle(E), loop(ES).

% internal events
handle(send_read(ID, PID)) :-
    '$node'(SELF), 
    send_with(SELF, PID, read(ID, PID, SELF)).
handle(send_value(ID, PID, X)) :-
    '$send'(PID, value(ID, PID, X)).
handle(send_get_module(NID, MID, MOD)) :-
    '$send'(NID, get_module(MID, MOD)).
handle(send_call(MID, IDX, ARGS, all)) :-
    '$node'(SELF),
    broadcast(SELF, call(MID, IDX, SELF, ARGS)).
handle(send_call(MID, IDX, ARGS, PEER)) :-
    PEER =\= all |
    '$resolve_peer'(PEER, PID),
    '$node'(SELF),
    send_with(SELF, PID, call(MID, IDX, SELF, ARGS)).
handle(send_lookup(NAME, PEER)) :-
    '$node'(SELF),
    send_with(SELF, 0, lookup(NAME, PEER, SELF)).
handle(send_addref(PID, ID)) :-
    '$send'(PID, addref(ID)).
handle(send_assign_port(PID, ID, X)) :-
    '$send'(PID, assign_port(ID, X)).
handle(send_idle_token(TOK)) :-
    get_arg(1, TOK, NODES),
    get_arg(2, TOK, INDEX),
    get_arg(INDEX, NODES, PID),
    '$send'(PID, idle_token(TOK)).
handle(send_idle_broadcast) :-
    '$send'(0, idle_broadcast).
% external events
handle(read(ID, SELF, PID)) :-
    '$get_exposed'(ID, X), 
    send_value(X, PID, ID, SELF).  % suspends on X
handle(value(ID, OWNER, X)) :- 
    '$assign_remote'(ID, OWNER, X).
handle(call(MID, IDX, NID, ARGS)) :-
    '$retrieve_module'(MID, NID, MOD),
    '$call'(MOD, IDX, ARGS).
handle(get_module(MID, MOD)) :-
    '$load_module'(MID, MOD).
handle(peer(PEER, NODE)) :-
    '$register_peer'(PEER, NODE).
handle(silence) :- '$silence'.
handle(terminate) :- halt(0).
handle(exit(REASON)) :-
    '$node'(SELF, PID),
    fmt:format(2, '[~q:~q] terminated - ~q~n', [PID, SELF, REASON],
        DONE),
    done(DONE, 1).
handle(addref(ID)) :-
    '$add_reference'(ID).
handle(assign_port(ID, X)) :-
    '$assign_port'(ID, X).
handle(idle_token(TOK)) :-
    '$set_idle_token'(TOK).
handle(idle) :-
    '$network_is_idle'.
handle(E) :-
    otherwise |
    error(unkown_event(E)).

done([], CODE) :- halt(CODE).

send_with(SELF, DEST, MSG) :- 
    data(SELF) | '$send'(DEST, MSG).

send_value(X, PID, ID, SELF) :-
    data(X) |
    '$drop_reference'(ID),
    '$send'(PID, value(ID, SELF, X)).

broadcast(SELF, CALL) :-
    get_global('$all_peers', ALL, none),
    broadcast(ALL, SELF, CALL).

broadcast(none, SELF, CALL) :-
    nodes(ALL),
    put_global('$all_peers', ALL, _),
    broadcast(ALL, SELF, CALL).
broadcast([], _, _).
broadcast([PID|ALL], SELF, CALL) :-
    data(SELF) | 
    '$send'(PID, CALL),
    broadcast(ALL, SELF, CALL).

% exported interface

nodes(ALL) :- '$send'(0, nodes(ALL)).
terminate :- '$send'(0, shutdown).
self(SELF) :- '$node'(SELF).
register(NAME) :- 
    '$node'(SELF), 
    send_with(SELF, 0, register(SELF, NAME)).
